#!/usr/bin/env python

import os
from os import path
from unittest import TestCase

from fastalite import fastalite, fastqlite, Opener

outdir = 'test_output'


class TestOpener(TestCase):

    def setUp(self):
        funcname = '_'.join(self.id().split('.')[-2:])
        self.fn = 'testfiles/ten.fasta'
        d = path.join(outdir, funcname)
        if not path.exists(d):
            os.makedirs(d)

        self.basename = path.join(outdir, funcname, path.basename(self.fn))

        with open(self.fn) as f:
            self.seqs = list(fastalite(f))

    def test_opener(self):
        with Opener()(self.fn) as infile:
            seqs = fastalite(infile)
            self.assertEqual(list(seqs), self.seqs)

    def test_gz(self):
        fname = self.basename + '.gz'
        with Opener('w')(fname) as outfile:
            for seq in self.seqs:
                outfile.write('>{}\n{}\n'.format(seq.description, seq.seq))

        with Opener()(fname) as infile:
            seqs = fastalite(infile)
            self.assertEqual(list(seqs), self.seqs)

    def test_bz2(self):
        fname = self.basename + '.bz2'
        with Opener('w')(fname) as outfile:
            for seq in self.seqs:
                outfile.write('>{}\n{}\n'.format(seq.description, seq.seq))

        with Opener()(fname) as infile:
            seqs = fastalite(infile)
            self.assertEqual(list(seqs), self.seqs)


class TestFastq(TestCase):

    def test_fastq1(self):
        with Opener()('testfiles/good.fastq') as infile:
            seqs = fastqlite(infile)
            self.assertEqual(len(list(seqs)), 20)

    def test_fastq2(self):
        # "'+' replaced with '-' on line 15"
        with Opener()('testfiles/bad1.fastq') as infile:
            seqs = fastqlite(infile)
            self.assertRaises(ValueError, list, seqs)

    def test_fastq3(self):
        # "final qual line missing"
        with Opener()('testfiles/bad2.fastq') as infile:
            seqs = fastqlite(infile)
            self.assertRaises(ValueError, list, seqs)

    def test_fastq4(self):
        # "first qual line missing one character"
        with Opener()('testfiles/bad3.fastq') as infile:
            seqs = fastqlite(infile)
            self.assertRaises(ValueError, list, seqs)



import sys
import gzip
from collections import namedtuple

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

try:
    import bz2
except ImportError as err:
    def bz2_open(filename, mode, *args, **kwargs):
        sys.exit(err)
else:
    bz2_open = bz2.open if hasattr(bz2, 'open') else bz2.BZ2File


class Opener(object):
    """Factory for creating file objects. Transparenty opens compressed
    files for reading or writing based on suffix (.gz and .bz2 only).

    Example::

        with Opener()('in.txt') as infile, Opener('w')('out.gz') as outfile:
            outfile.write(infile.read())
    """

    def __init__(self, mode='r', *args, **kwargs):
        self.mode = mode
        self.args = args
        self.kwargs = kwargs
        self.writable = 'w' in self.mode

    def __call__(self, obj):
        if obj is sys.stdout or obj is sys.stdin:
            return obj
        elif obj == '-':
            return sys.stdout if self.writable else sys.stdin
        else:
            openers = {'bz2': bz2_open, 'gz': gzip.open}
            __, suffix = obj.rsplit('.', 1)
            # in python3, both bz2 and gz libraries default to binary input and output
            mode = self.mode
            if sys.version_info.major == 3 and suffix in openers \
               and mode in {'w', 'r'}:
                mode += 't'
            opener = openers.get(suffix, open)
            return opener(obj, mode=mode, *self.args, **self.kwargs)


def fastalite(handle):
    """Return a sequence of namedtuple objects from a fasta file with
    attributes (id, description, seq) given open file-like object
    ``handle``

    """

    Seq = namedtuple('Seq', ['id', 'description', 'seq'])

    header, seq = '', []
    for line in handle:
        if line.startswith('>'):
            if header:
                yield Seq(header.split()[0], header, ''.join(seq))
            header, seq = line[1:].strip(), []
        else:
            seq.append(line.strip())

    if header and seq:
        yield Seq(header.split()[0], header, ''.join(seq))


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def fastqlite(handle):
    """Return a sequence of namedtuple objects from a fastq file with
    attributes (id, description, seq, qual) given open file-like
    object ``handle``. This parser assumes that lines corresponding to
    sequences and quality scores are not wrapped. Raises
    ``ValueError`` for malformed records.

    See https://doi.org/10.1093/nar/gkp1137 for a discussion of the
    fastq format.

    """

    Seq = namedtuple('Seq', ['id', 'description', 'seq', 'qual'])
    for i, chunk in enumerate(grouper(handle, 4, '')):
        description, seq, plus, qual = chunk
        seq, qual = seq.strip(), qual.strip()

        checks = [description.startswith('@'), seq,
                  plus.startswith('+'), qual, len(seq) == len(qual)]

        if not all(checks):
            raise ValueError('Malformed record around line {}'.format(i * 4))

        description = description[1:].strip()
        yield Seq(description.split()[0], description, seq, qual)



"""Command line interface to the fastlite package

"""

import sys
import argparse
from .fastalite import fastalite, fastqlite, Opener
from . import __version__


def count(seqs):
    for i, seq in enumerate(seqs, 1):
        pass
    print(i)


def names(seqs):
    for seq in seqs:
        print((seq.id))


def lengths(seqs):
    for seq in seqs:
        print(('{}\t{}'.format(seq.id, len(seq.seq))))


def main(arguments):

    actions = {'count': count, 'names': names, 'lengths': lengths}

    parser = argparse.ArgumentParser(
        prog='fastalite', description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('action', choices=list(actions.keys()),
                        help="Action to perform")
    parser.add_argument('infile', help="Input file",
                        metavar='infile.{fasta,fastq}[{.gz,.bz2}]',
                        type=Opener('r'))
    parser.add_argument('-V', '--version', action='version',
                        version=__version__,
                        help='Print the version number and exit')
    args = parser.parse_args(arguments)

    with args.infile as f:
        readfun = fastalite if 'fasta' in f.name else fastqlite
        seqs = readfun(args.infile)
        try:
            fun = actions[args.action]
            fun(seqs)
        except ValueError as err:
            sys.stderr.write('Error: {}\n'.format(str(err)))
            return 1

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import os
import subprocess
from setuptools import setup, find_packages

subprocess.call(
    ('mkdir -p fastalite/data && '
     'git describe --tags --dirty > fastalite/data/ver.tmp'
     '&& mv fastalite/data/ver.tmp fastalite/data/ver '
     '|| rm -f fastalite/data/ver.tmp'),
    shell=True, stderr=open(os.devnull, "w"))

from fastalite import __version__

setup(
    author='Noah Hoffman',
    author_email='noah.hoffman@gmail.com',
    description='Simplest possible fasta parser',
    url='https://github.com/nhoffman/fastalite',
    name='fastalite',
    packages=find_packages(),
    package_dir={'fastalite': 'fastalite'},
    package_data={'fastalite': ['data/ver']},
    version=__version__,
    test_suite='tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

    ],
)
